import os.path
import signal
import collections

from twisted.names import dns
from twisted.names.authority import FileAuthority
from twisted.names.client import Resolver
from twisted.internet import inotify
from twisted.internet import defer
from twisted.python import filepath
from IPy import IP


def extract_zones_from_status_file(status_path):
    """ Parses a openvpn status file and extracts the list of connected clients
        and there ip address """
    with open(status_path, 'r') as status_file:
        mode = None
        mode_changes = {
            'OpenVPN CLIENT LIST': 'clients',
            'ROUTING TABLE': 'routes',
            'GLOBAL STATS': None
        }
        clients = {}
        skip_next_lines = 0
        for status_line in status_file:
            if skip_next_lines > 0:
                skip_next_lines -= 1
                continue
            status_line = status_line.strip()
            if status_line in mode_changes:
                mode = mode_changes[status_line]
                skip_next_lines = 1
                if mode == 'clients':
                    skip_next_lines += 1
                continue
            if mode == 'clients':
                clients[status_line.split(',')[0]] = []
            if mode == 'routes':
                address, client = status_line.split(',')[0:2]
                if '/' in address:  # subnet
                    continue
                try:
                    address = IP(address)
                except ValueError:  # cached route ...
                    continue
                if address.len() > 1:  # subnet
                    continue
                if client not in clients:
                    raise ValueError('Error in status file')
                clients[client].append(address)
        return clients


class InMemoryAuthority(FileAuthority):
    """ In memory authority class - handles the data of one zone"""
    def __init__(self, data=None):
        FileAuthority.__init__(self, data)

    def loadFile(self, data):
        if type(data) is tuple and len(data) == 2:
            self.setData(*data)

    def setData(self, soa, records):
        """ set authority data

            :param twisted.names.dns.Record_SOA soa: SOA record for this zone.
                you must add the soa to the records list yourself!!
            :param dict records: dictionary with record entries for this
                domain."""
        if soa == self.soa or self.changed(soa, records) is False:
            if type(soa) is tuple:
                print('skip unchanged update for zone {0}'.format(soa[0]))
            return False
        self.soa = soa
        self.records = records
        return True

    def changed(self, soa, records):
        """ Checks whether the new record list differs from the old one"""
        if self.records is None:  # previously set data
            return True
        if len(self.records) != len(records):
            return True
        for name in self.records:
            if name not in records:
                return True
            if len(self.records[name]) != len(records[name]):
                return True
            for record in self.records[name]:
                for new_record in records[name]:
                    if new_record == record:
                        break
                    if new_record.__class__ is dns.Record_SOA and \
                            record.__class__ is dns.Record_SOA and \
                            new_record.mname == record.mname and \
                            new_record.rname == record.rname and \
                            new_record.refresh == record.refresh and \
                            new_record.retry == record.retry and \
                            new_record.expire == record.expire and \
                            new_record.minimum == record.minimum:
                        break
                else:
                    return True
        return False


AuthorityTuple = collections.namedtuple('AuthorityTuple', ('forward',
                                        'backward4', 'backward6'))


class OpenVpnAuthorityHandler(list):
    def __init__(self, config):
        self.config = config
        self.send_notify = False
        # authorities for the data itself:
        self.authorities = {}
        for instance in self.config.instances:
            self.authorities[instance] = AuthorityTuple(
                forward=InMemoryAuthority(),
                backward4=InMemoryAuthority(),
                backward6=InMemoryAuthority()
            )
            self.append(self.authorities[instance].forward)
            if self.config.instances[instance].subnet4:
                self.append(self.authorities[instance].backward4)
            if self.config.instances[instance].subnet6:
                self.append(self.authorities[instance].backward6)
        # load data:
        self.loadInstances()
        # watch for file changes:
        signal.signal(signal.SIGUSR1, self.handle_signal)
        notifier = inotify.INotify()
        notifier.startReading()
        for instance in self.config.instances.values():
            notifier.watch(filepath.FilePath(instance.status_file),
                           callbacks=[self.status_file_changed])
        print('Serving {0} zones: {1}'.format(len(self),
              ', '.join(map(lambda z: z.soa[0], self))))

    def loadInstances(self):
        """ (re)load data of all instances"""
        for instance in self.config.instances:
            self.loadInstance(self.config.instances[instance])

    def loadInstance(self, instance):
        clients = extract_zones_from_status_file(instance.status_file)
        self.build_zone_from_clients(instance, clients)

    @staticmethod
    def create_record_base(zone_name, soa, initial_data):
        if zone_name is None:
            return {}
        records = {}
        records.setdefault(zone_name, []).append(soa)
        for name, record in initial_data:
            if name == '@':
                name = zone_name
            elif not name.endswith('.'):
                name = name + '.' + zone_name
            records.setdefault(name, []).append(record)
        return records

    def build_zone_from_clients(self, instance, clients):
        """ Basic zone generation (uses only the client list),
            additional data like SOA information must be passed
            as keyword option """
        soa = dns.Record_SOA(
            mname=instance.mname,
            rname=instance.rname,
            serial=int(os.path.getmtime(instance.status_file)),
            refresh=instance.refresh,
            retry=instance.retry,
            expire=instance.expire,
            minimum=instance.minimum,
        )
        forward_records = self.create_record_base(instance.name, soa,
                                                  instance.forward_records)
        backward4_records = self.create_record_base(instance.subnet4, soa,
                                                    instance.backward4_records)
        backward6_records = self.create_record_base(instance.subnet6, soa,
                                                    instance.backward6_records)
        for client, addresses in clients.items():
            if instance.suffix is not None:
                if instance.suffix == '@':
                    client += '.' + instance.name
                else:
                    client += '.' + instance.suffix
            client = client.lower()
            for address in addresses:
                reverse = IP(address).reverseName()[:-1]
                if address.version() == 4:
                    forward_records.setdefault(client, []) \
                        .append(dns.Record_A(str(address)))
                    backward4_records.setdefault(reverse, []) \
                        .append(dns.Record_PTR(client))
                elif address.version() == 6:
                    forward_records.setdefault(client, []) \
                        .append(dns.Record_AAAA(str(address)))
                    backward6_records.setdefault(reverse, []) \
                        .append(dns.Record_PTR(client))
        # push data to authorities:
        authority = self.authorities[instance.name]
        if authority.forward.setData((instance.name, soa), forward_records):
            self.notify(instance, instance.name)
        if instance.subnet4:
            if authority.backward4.setData((instance.subnet4, soa), backward4_records):
                self.notify(instance, instance.subnet4)
        if instance.subnet6:
            if authority.backward6.setData((instance.subnet6, soa), backward6_records):
                self.notify(instance, instance.subnet6)

    def handle_signal(self, a, b):
        self.loadInstances()

    def status_file_changed(self, ignored, filepath, mask):
        instance = None
        for one_instance in self.config.instances.values():
            if one_instance.status_file == filepath.path:
                instance = one_instance
                break
        if instance is None:
            print('unknown status file: {0}'.format(filepath.path))
            return
        print('{0} changed ({1}), rereading instance {2}'.format(filepath,
              ','.join(inotify.humanReadableMask(mask)), instance.name))
        self.loadInstance(instance)

    def notify(self, instance, name):
        if self.send_notify is not True:
            return
        for server in instance.notify:
            print('Notify {0} new data for zone {1}'.format(server[0], name))
            r = NotifyResolver(servers=[server])
            r.sendNotify(name)

    def start_notify(self):
        self.send_notify = True
        for instance in self.config.instances.values():
            self.notify(instance, instance.name)
            if instance.subnet4:
                self.notify(instance, instance.subnet4)
            if instance.subnet6:
                self.notify(instance, instance.subnet6)


class NotifyResolver(Resolver):
    def sendNotify(self, zone):
        protocol = self._connectedProtocol()

        id = protocol.pickID()

        m = dns.Message(id, opCode=dns.OP_NOTIFY)
        m.queries = [dns.Query(zone, dns.SOA, dns.IN)]

        try:
            protocol.writeMessage(m, self.servers[0])
        except:
            return defer.fail()

        resultDeferred = defer.Deferred()
        cancelCall = protocol.callLater(10, protocol._clearFailed, resultDeferred, id)
        protocol.liveMessages[id] = (resultDeferred, cancelCall)

        d = resultDeferred

        def cbQueried(result):
            protocol.transport.stopListening()
            return result
        d.addBoth(cbQueried)
        return d



from setuptools import setup  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

if not path.exists(path.join(here, 'README.rst')):
    try:
        import subprocess
        subprocess.call(['pandoc',
                         '--from=markdown', '--to=rst',
                         path.join(here, 'README.md'),
                         '--output', path.join(here, 'README.rst')])
    except:
        import os
        os.copy(path.join(here, 'README.md'), path.join(here, 'README.rst'))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

import version


setup(
    name='openvpn2dns',
    version=version.STRING,

    description='DNS server in python serving openvpn status files as dns zones',
    long_description=long_description,

    url='https://github.com/mswart/openvpn2dns',

    author='Malte Swart',
    author_email='mswart@devtation.de',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: System Administrators',
        'Topic :: Internet :: Name Service (DNS)',
        'Topic :: System :: Networking',
        'Topic :: Communications',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    keywords='openvpn twisted dns',

    install_requires=[
        #'Twisted >= 10', diabled as only twisted-names is needed
        'IPy >= 0.73'
    ],
    py_modules=('config', 'openvpnzone', 'version'),
    scripts=('openvpn2dns', )
)



from __future__ import print_function

import os.path
import warnings
import pwd
import grp

from twisted.names import dns
from IPy import IP


class ConfigurationError(Exception):
    pass


class MissingSectionError(ConfigurationError):
    pass


class InstanceRedifinitionError(ConfigurationError):
    pass


class ConfigurationWarning(UserWarning):
    pass


class UnusedOptionWarning(ConfigurationWarning):
    pass


class OptionRedifinitionWarning(ConfigurationWarning):
    pass


def extract_status_file_path(config_path):
    """ Extracts the path to the openvpn status file from a openvpn config.

        :raises EnvironmentError: if the config does not contain a status file
            entry
        :return: the absolute path to the status file"""
    import re
    status_search = re.compile(r'^\s*(?P<option>(status|server|server-ipv6))\s+(?P<value>[^#]+)(#.*)?$')

    status_file = None
    subnet4 = None
    subnet6 = None

    with open(config_path, 'r') as conf_file:
        for conf_line in conf_file:
            match = status_search.match(conf_line)
            if not match:
                continue
            if match.group('option') == 'status':
                status_file = os.path.abspath(os.path.join(config_path, '..',
                                              match.group('value').strip()))
            elif match.group('option') == 'server':
                subnet4 = match.group('value').strip()
            elif match.group('option') == 'server-ipv6':
                subnet6 = match.group('value').strip()

    if not status_file:
        raise EnvironmentError('You must specify a status entry!')

    return (status_file, subnet4, subnet6)


class SetSingleValueMixin:
    def set_single_option(self, name, value, convert=None):
        old_value = getattr(self, name, object())
        if old_value is not None:
            warnings.warn('Overwrite old {0} option value ({1})'
                          .format(name, old_value), OptionRedifinitionWarning,
                          stacklevel=2)
        if convert is not None:
            try:
                value = convert(value)
            except Exception as e:
                raise ConfigurationError(e.message)
        setattr(self, name, value)


class OpenVpnInstance(object, SetSingleValueMixin):
    def __init__(self, name):
        self.name = name
        self.status_file = None
        self.notify = []
        self.rname = None
        self.mname = None
        self.refresh = None
        self.retry = None
        self.expire = None
        self.minimum = None
        self.forward_records = []
        self.backward4_records = []
        self.backward6_records = []
        self.suffix = None
        self.subnet4 = None
        self.subnet6 = None


class ConfigParser(object, SetSingleValueMixin):
    """ Parser and data storage for configuration information.
        The file format uses the INI syntax but with multiple use of option
        names per section therefore the config module is not usable.

        :param str filename: file path to configuration file"""
    def __init__(self, filename=None):
        self.listen_addresses = []
        self.daemon = None
        self.drop = None
        self.user = None
        self.group = None
        self.log = None
        self.pidfile = None
        self.reactor = None
        self.instances = {}
        if filename:
            self.read_file(filename)

    @staticmethod
    def read_data(filename):
        """ Reads the given file name. It assumes that the file has a INI file
            syntax. The parser returns the data without comments and fill
            characters. It supports multiple option with the same name per
            section but not multiple sections with the same name.

            :param str filename: path to INI file
            :return: a dictionary - the key is the section name value, the
                option is a array of (option name, option value) tuples"""
        sections = {}
        with open(filename) as f:
            section = None
            options = None
            for line in f:
                # ignore comments:
                if line.startswith('#'):
                    continue
                line = line.strip()
                if not line:
                    continue
                # handle section header:
                if line.startswith('[') and line.endswith(']'):
                    if section:  # save old section data
                        sections[section] = options
                    section = line[1:-1]
                    options = []
                    continue
                if section is None:
                    warnings.warn('Option without sections: {0}'.format(line),
                                  UnusedOptionWarning, stacklevel=2)
                    continue
                option, value = line.split('=', 1)
                options.append((option.strip(), value.strip()))
            if section:  # save old section data
                sections[section] = options
        return sections

    def read_file(self, file_name):
        """ Read and parse the configuration file"""
        self.parse_data(self.read_data(file_name))

    @staticmethod
    def parse_boolean(value):
        if value.lower() in ['true', 't', 'yes', 'y', '1']:
            return True
        if value.lower() in ['false', 'f', 'no', 'n', '0']:
            return False
        raise ValueError('Could not parse boolean value')

    @staticmethod
    def parse_filename(value):
        if not os.path.isfile(value):
            raise ValueError('Could not found config file {0}'.format(value))
        return value

    @staticmethod
    def parse_userid(value):
        try:
            return int(value)
        except ValueError:
            pass
        try:
            return pwd.getpwnam(value)[2]
        except KeyError:
            raise ValueError('Unknown user name or id: {0}'.format(value))

    @staticmethod
    def parse_groupid(value):
        try:
            return int(value)
        except ValueError:
            pass
        try:
            return grp.getgrnam(value)[2]
        except KeyError:
            raise ValueError('Unknown group name or id: {0}'.format(value))

    @staticmethod
    def parse_net(value):
        return IP(value.replace(' ', '/')).reverseName()[:-1]

    def parse_data(self, data):
        """ Parse configuration data

            :param dict data: data extracted from :meth:`read_data`
            :raises ConfigurationError: missing information"""
        # store data for other methods:
        self.data = data
        # handle main options:
        if 'options' not in data:
            raise MissingSectionError('Missing options section')
        for option, value in data['options']:
            if option == 'listen':
                self.add_listen_address(value)
            elif option == 'instance':
                self.parse_instance(value)
            elif option == 'reactor':
                self.set_single_option('reactor', value)
            elif option == 'log':
                self.set_single_option('log', value)
            elif option == 'daemon':
                self.set_single_option('daemon', value, self.parse_boolean)
            elif option == 'drop-privileges':
                self.set_single_option('drop', value, self.parse_boolean)
            elif option == 'user':
                self.set_single_option('user', value, self.parse_userid)
            elif option == 'group':
                self.set_single_option('group', value, self.parse_groupid)
            elif option == 'pidfile':
                self.set_single_option('pidfile', value, self.parse_boolean)
            else:
                warnings.warn('Unknown option {0} in options section'
                              .format(option), UnusedOptionWarning,
                              stacklevel=2)

    def add_listen_address(self, listen):
        """ Add a listen information

            :param str listen: listen information (name, ip, port)"""
        address, port = listen.split(':', 1)
        self.listen_addresses.append((address, int(port)))

    def parse_instance(self, name):
        """ register one openvpn status instance

            :param str name: name for this instance (used as zone name)"""
        if name in self.instances:
            raise InstanceRedifinitionError('Instance {0} already defined'
                                            .format(name))
        instance = OpenVpnInstance(name)
        if name not in self.data:
            raise MissingSectionError('section for instance {0}'.format(name))
        for option, value in self.data[name]:
            # openvpn server information:
            if option == 'server_config':
                status, subnet4, subnet6 = extract_status_file_path(value)
                instance.set_single_option('status_file', status)
                if subnet4:
                    instance.set_single_option('subnet4', subnet4, self.parse_net)
                if subnet6:
                    instance.set_single_option('subnet6', subnet6, self.parse_net)
            elif option == 'status_file':
                instance.set_single_option('status_file', os.path.abspath(value))
            elif option == 'subnet4':
                instance.set_single_option('subnet4', value, self.parse_net)
            elif option == 'subnet6':
                instance.set_single_option('subnet6', value, self.parse_net)
            # slave name server notifies:
            elif option == 'notify':
                instance.notify.append((value, 53))
            elif option == 'suffix':
                instance.suffix = value
            # SOA entries:
            elif option in ('rname', 'mname', 'refresh', 'retry', 'expire',
                            'minimum'):
                if option == 'rname':
                    value = value.replace('@', '.', 1)
                if getattr(instance, option) is not None:
                    warnings.warn('Overwrite SOA value {0}'.format(option),
                                  OptionRedifinitionWarning, stacklevel=2)
                setattr(instance, option, value)
            # additional entries:
            elif option in ['add_entries',
                            'add_forward_entries', 'add_backward_entries',
                            'add_backward4_entries', 'add_backward6_entries']:
                if value not in self.data:
                    raise MissingSectionError('Referencing unknown section {0}'
                                              .format(value))
                records = self.parse_entry_section(self.data[value])
                type = option[4:-8]
                if type in ('', 'forward'):
                    instance.forward_records += records
                if type in ('', 'backward', 'backward4'):
                    instance.backward4_records += records
                if type in ('', 'backward', 'backward6'):
                    instance.backward6_records += records
            else:
                warnings.warn('Unknown option {0} in section {1}'.format(option,
                              name), UnusedOptionWarning, stacklevel=2)
        self.instances[name] = instance
        return instance

    @staticmethod
    def parse_entry_section(options, name=None):
        records = []
        for entry, value in options:
            parts = value.split(' ')
            record = getattr(dns, 'Record_%s' % parts[0], None)
            if not record:
                raise NotImplementedError("Record type %r not supported" % type)
            records.append((entry, record(*parts[1:])))
        return records
